/*  Copyright (C) 1996-2001  Martin Vicente			      -*- c -*-
    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/****  Project : eXtended C Library (XCLIB)
    *  Purpose : GENERIC LIST
    *  Language: ANSI C
    *  File    : demo_list-4.c
    *  Author  : Martin Vicente <martin@vicente.org>
    *
    ****/


/*****************************************************************************\

 [English]

 <xc/list> use with type checking.

 [French]

 Utilisation de <xc/list> avec contrle de type.

\*****************************************************************************/


#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include <xc/strlib.h>
#include <xc/xalloc.h>
#include <xc/list.h>

typedef struct ObjectRec *Object;

def_LIST(OBJECTS,, Object);

#define OBJECTS	(*OBJECTS)	/* optional */

struct ObjectRec
{
    int   number;
    char *name;
};

static Object createObject(const char *name);
static void   destroyObject(Object object);

static int print_object(ObjectId id, void *data);
static int compare_objects(ObjectId id1, ObjectId id2, void *data);
static int test_object(ObjectId id, void *name);
static int destroy_object(ObjectId id, void *data);

#define COMM(en, fr)	__xcint(comm)(en, fr, NULL)
#define WAIT()		__xcint(wait)()

extern void __xcint(comm)(const char *en, ...);	/* DON'T USE IT! */
extern void __xcint(wait)(void);		/* DON'T USE IT! */

int main()
{
    ObjectList objects;
    ObjectId   id;

    COMM("Creating list of objects",
	 "Cration de la liste d'objets");

    objects = OBJECTS.create(0);

    OBJECTS.add(objects, createObject("Black"));
    OBJECTS.add(objects, createObject("Red"));
    OBJECTS.add(objects, createObject("Green"));
    OBJECTS.add(objects, createObject("Yellow"));
    OBJECTS.add(objects, createObject("Blue"));
    OBJECTS.add(objects, createObject("Magenta"));
    OBJECTS.add(objects, createObject("Cyan"));
    OBJECTS.add(objects, createObject("White"));

    OBJECTS.action(objects, print_object, NULL);

    WAIT();

    COMM("Sorting objects",
	 "Tri des objets");

    OBJECTS.sort(objects, compare_objects, NULL);

    OBJECTS.action(objects, print_object, NULL);

    WAIT();

    COMM("Searching and removing an object from list",
	 "Recherche et extraction d'un objet de la liste");

    if ((id = OBJECTS.search(objects, test_object, "Green")) != NULL)
    {
	Object object;

	printf("\tRemoving Object: %s\n", id->item->name);

	object = OBJECTS.remove(id);

	printf("\tObject Removed:  %s\n", object->name);

	putchar('\n');

	OBJECTS.action(objects, print_object, NULL);
    }

    WAIT();

    COMM("Creating an table from list",
	 "Production d'un tableau  partir de la liste");

    {
	Object table = OBJECTS.table(objects, sizeof (struct ObjectRec));
	size_t count = OBJECTS.count(objects);
	size_t i;

	for (i = 0; i < count; ++i) printf("\ttable[%u].name = %s\n", i, table[i].name);

	for (i = 0; i < count; ++i) table[i].name = strupp(xstrdup(table[i].name));

	putchar('\n');

	for (i = 0; i < count; ++i) printf("\ttable[%u].name = %s\n", i, table[i].name);

	putchar('\n');

	OBJECTS.action(objects, print_object, NULL);

	for (i = 0; i < count; ++i) xfree(table[i].name);

	xfree(table);
    }

    WAIT();

    COMM("Destroying list",
	 "Destruction de la liste d'objets");

    OBJECTS.action(objects, destroy_object, NULL);
    OBJECTS.destroy(objects);

    return 0;
}

static Object createObject(const char *name)
{
    static int number = 0;

    Object object;

    assert(name != NULL);

    object = xnew(Object);

    object->number = ++number;
    object->name   = xstrdup(name);

    return object;
}

static void destroyObject(Object object)
{
    assert(object != NULL);

    xfree(object->name);

    xfree(object);
}

static int print_object(ObjectId id, __unused void *data)
{
    Object object;

    object = OBJECTS.item(id);	/* Ces deux mthodes sont aussi */
    object = id->item;		/* valable l'une que l'autre.   */

    printf("\tObject %d - %s\n", object->number, object->name);

    return 0;
}

static int compare_objects(ObjectId id1, ObjectId id2, __unused void *data)
{
    /* Sorting objects by his name in lexicographical order */

    return strcmp(id1->item->name, id2->item->name);
}

static int test_object(ObjectId id, void *name)
{
    assert(name != NULL);

    return strcmp(id->item->name, name) == 0;
}

static int destroy_object(ObjectId id, __unused void *data)
{
    destroyObject(id->item);

    return 0;
}


/* End of file */
