/*  XCLIB - a set of C functions for general purposes		      -*- c -*-
    Copyright (C) 1996-2001  Martin Vicente

    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*\
  |  Martin Vicente, 1996-2001
  |
  | Project : XCLIB (eXtended C Library)
  | Purpose : SIMPLIFIED TIMER INTERFACE
  | Version : 0.3
  | Language: ANSI C
  | File    : timer/timer.c
  | Author  : Martin VICENTE <martin@vicente.org>
\*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*/


#define _XCLIB_INTERNALS

#include <sys/time.h>

#include <unistd.h>

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

#include <xc/defines.h>
#include <xc/types.h>

#include "xc/timer.h"

#include "global.h"

#define EQ(t1, t2) ((t1).tv_sec == (t2).tv_sec && (t1).tv_usec == (t2).tv_usec)

#define FIXUP_TIMEVAL(t)						      \
_MBEGIN									      \
 while ((t).tv_usec <= MILLION) (t).tv_usec += MILLION, --(t).tv_sec;	      \
 while ((t).tv_usec >= MILLION) (t).tv_usec -= MILLION, ++(t).tv_sec;	      \
 if ((t).tv_usec < 0 && (t).tv_sec > 0) (t).tv_usec += MILLION, --(t).tv_sec; \
 if ((t).tv_usec > 0 && (t).tv_sec < 0) (t).tv_usec -= MILLION, ++(t).tv_sec; \
_MEND

static const timeval_t zero = { 0, 0 };

const struct timer *timer_ctl(const struct timer *timer, unsigned int flags)
{
    struct timer *t = (struct timer *)timer;

    if (flags & timerNEW)
    {
	if (t != NULL) return NULL;

	t = malloc(sizeof (struct timer));

	if (t == NULL) return NULL;

	t->now   = zero;
	t->start = zero;
	t->last  = zero;
	t->count = 0;
    }

    if (flags & timerSET)
    {
	if (t == NULL) return NULL;

	if (gettimeofday(&t->now, NULL) == -1) return NULL;

	FIXUP_TIMEVAL(t->now);

	t->start = t->now;
	t->last  = zero;
	t->count = 0;
    }

    if (flags & timerTOP)
    {
	register struct timeval prev;

	if (t == NULL) return NULL;

	if (EQ(t->now, zero) || EQ(t->start, zero)) return NULL;

	prev = t->now;

	if (gettimeofday(&t->now, NULL) == -1) return NULL;

	FIXUP_TIMEVAL(t->now);

	t->last.tv_sec  = t->now.tv_sec  - prev.tv_sec;
	t->last.tv_usec = t->now.tv_usec - prev.tv_usec;

	FIXUP_TIMEVAL(t->last);

	t->count++;
    }

    if (flags & timerPUT)
    {
	timeval_t tt;

	register unsigned long int n;

	if (t == NULL) return NULL;

	n = t->count;

	if (n > 0) printf("#%lu ", n);

	FIXUP_TIMEVAL(t->last);

	printf(
	    "%3ld s %3ld ms %3ld us",
	    (long int)t->last.tv_sec,
	    (long int)t->last.tv_usec / 1000,
	    (long int)t->last.tv_usec % 1000);

	tt.tv_sec  = t->now.tv_sec  - t->start.tv_sec;
	tt.tv_usec = t->now.tv_usec - t->start.tv_usec;

	FIXUP_TIMEVAL(tt);

	printf(
	    " | TOTAL %3ld s %3ld ms %3ld us",
	    (long int)tt.tv_sec,
	    (long int)tt.tv_usec / 1000,
	    (long int)tt.tv_usec % 1000);

	if (n > 1)
	{
	    timeval_t av;

	    float usec;

	    unsigned int nsec;

	    av.tv_sec  = tt.tv_sec / n;

	    usec = (float)(MILLION * (tt.tv_sec % n) + tt.tv_usec) / (float)n;

	    av.tv_usec = usec;

	    nsec = 1000 * (usec - av.tv_usec);

	    printf(
		" | AVERAGE %3ld s %3ld ms %3ld us %3u ns",
		(long int)av.tv_sec,
		(long int)av.tv_usec / 1000,
		(long int)av.tv_usec % 1000,
		nsec);
	}

	fputc('\n', stdout);
    }

    if (flags & timerDEL)
    {
	if (t == NULL) return NULL;

	free(t);

	t = NULL;
    }

    return t;
}


/* End Of File */
